import matplotlib as mpl
from ._setup import setup
from ._format import add_colorbar, format_datetime_axis
from ._draw import draw_arrows, draw_earth, draw_multicolored_line


mplc = mpl.colors
plt = mpl.pyplot
setup()



__all__ = ["interpol"]

import astropy.units as u
import numpy as np
from scipy.interpolate import interp1d
from .timing import sampling_period
from .moving import move_avg


def interpol(y, x, xout, window=None, kind="linear"):
    r"""Interpolate a signal onto a new temporal grid.

    This function mimics IDL's interpol routine. When downsampling, the
    window argument can be set to apply a moving average over the
    original temporal grid. Additionally, astropy.Quantity are preserved
    from this function.
    """

    if not (
        (
            np.issubdtype(x.dtype, np.datetime64)
            and np.issubdtype(xout.dtype, np.datetime64)
        )
    ):
        raise NotImplementedError("Time arrays must be datetime64 dtype.")

    interp_kw = dict(kind=kind, bounds_error=False, fill_value=np.nan)
    if window is not None:
        w = (sampling_period(xout) / sampling_period(x)).decompose()
        assert w >= 1, "xout has higher resolution!"
        if window == "box":
            w = np.int64(max(w, 1))

    x = x.astype("f8")
    xout = xout.astype("f8")

    if len(y.shape) == 1:
        yin = move_avg(y, (w,), window=window) if window is not None else y
        yout = interp1d(x, yin, **interp_kw)(xout)
    elif len(y.shape) == 2:
        yout = np.empty((len(xout), N := y.shape[1]))
        for i in range(N):
            yin = (
                move_avg(y[:, i], (w,), window=window)
                if window is not None
                else y[:, i]
            )
            yout[:, i] = interp1d(x, yin, **interp_kw)(xout)
    else:
        raise NotImplementedError("Max dimensionality is 2.")

    if isinstance(y, u.Quantity):
        yout *= u.Unit(y.unit)

    return yout



__all__ = [
    "add_colorbar",
    "format_datetime_axis",
]

import matplotlib.dates as mdates
from mpl_toolkits.axes_grid1 import make_axes_locatable


def add_colorbar(axis, where="right", pad=0.05, size="2%"):
    r"""Adds colorbar next to an axis"""
    divider = make_axes_locatable(axis)
    return divider.append_axes(where, size=size, pad=pad)


def format_datetime_axis(axis):
    locator = mdates.AutoDateLocator()
    formatter = mdates.ConciseDateFormatter(locator)
    axis.xaxis.set_major_formatter(formatter)



__all__ = [
    "draw_arrows",
    "draw_earth",
    "draw_multicolored_line",
]

import numpy as np
from matplotlib.collections import LineCollection
from matplotlib.patches import FancyArrowPatch
from matplotlib.colors import Normalize


def draw_arrows(axis, x, y, N=None, color="k"):
    r"""Draws arrows along a path on the axis
    """
    N = len(x) // 5 if N is None else N
    d = len(x) // (N + 1)

    idx = np.arange(d, len(x), d)
    for i in idx:
        ar = FancyArrowPatch(
            (x[i - 1], y[i - 1]),
            (x[i], y[i]),
            arrowstyle="->",
            mutation_scale=20,
            color=color,
        )
        axis.add_patch(ar)


def draw_earth(axis, R=1, N=50, zorder=999):
    r"""Draws the Earth on the axis (with shading indicating day/night)
    """
    # Nightside
    theta = np.linspace(np.pi / 2, 3 * np.pi / 2, N)
    Xn = R * np.cos(theta)
    Yn = R * np.sin(theta)
    Xn = np.append(Xn, Xn[0])
    Yn = np.append(Yn, Yn[0])
    # Dayside
    theta = np.linspace(-np.pi / 2, np.pi / 2, N)
    Xd = R * np.cos(theta)
    Yd = R * np.sin(theta)
    Xd = np.append(Xd, Xd[0])
    Yd = np.append(Yd, Yd[0])
    # Plot
    axis.plot(Xd, Yd, "-k", zorder=zorder)
    axis.plot(Xn, Yn, "-k", zorder=zorder)
    axis.fill(Xd, Yd, color="w")
    axis.fill(Xn, Yn, color="k")


def draw_multicolored_line(axis, x, y, c, cmap="jet",
                           vmin=None, vmax=None, set_lim=False, **kwargs):
    r"""Draws a line with colors on a scale determined by c
    """

    vmin = c.min() if vmin is None else vmin
    vmax = c.max() if vmax is None else vmax

    points = np.array([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)
    norm = Normalize(vmin, vmax)
    lc = LineCollection(segments, cmap=cmap, norm=norm, **kwargs)
    lc.set_array(c)
    line = axis.add_collection(lc)
    if set_lim:
        Lx = x.max() - x.min()
        Ly = y.max() - y.min()
        axis.set_xlim(x.min() - 0.1 * Lx, x.max() + 0.1 * Lx)
        axis.set_ylim(y.min() - 0.1 * Ly, y.max() + 0.1 * Ly)

    return line



__all__ = ["setup"]

import tempfile
import atexit
import shutil
import os
import matplotlib.pyplot as plt
import matplotlib as mpl
from importlib.resources import files


def _checkdep_usetex(s):
    return s and shutil.which("tex")


def setup(tex=False, cache=False, **kw):
    plt.style.use(files("tvolib.mpl_utils") / "data" / "mpl_utils.mplstyle")
    plt.rc("text", usetex=_checkdep_usetex(tex))
    if cache:
        # Quick fix for matplotlib's tex cache in multiprocessing
        mpldir = tempfile.mkdtemp()
        atexit.register(shutil.rmtree, mpldir)
        umask = os.umask(0)
        os.umask(umask)
        os.chmod(mpldir, 0o777 & ~umask)
        os.environ["HOME"] = mpldir
        os.environ["MPLCONFIGDIR"] = mpldir

        class TexManager(mpl.texmanager.TexManager):
            texcache = os.path.join(mpldir, "tex.cache")

        mpl.texmanager.TexManager = TexManager



__all__ = ["move_avg", "move_rms", "move_std"]

import numpy as np
from functools import reduce
from astropy.convolution import (
    Gaussian1DKernel,
    Gaussian2DKernel,
    CustomKernel,
    Box1DKernel,
    convolve,
)


def move_avg(x, w, window="box"):
    r"""Calculate the moving average of a signal."""

    if not isinstance(w, (tuple, list)):
        raise NotImplementedError("w must be a tuple or list.")
    if (ndim := x.ndim) != len(w := np.array(w)):
        raise NotImplementedError(
            "Length of w-tuple must be the same as x dimensions."
        )

    if ndim == 1:
        if window == "gauss":
            kernel = Gaussian1DKernel(*(w / 6))
        else:
            kernel = Box1DKernel(*w)
    elif ndim == 2:
        if window == "gauss":
            kernel = Gaussian2DKernel(*(w / 6), theta=np.pi / 2)
        else:
            kernel = CustomKernel(
                reduce(np.outer, (Box1DKernel(_w) for _w in w)).reshape(w)
            )
    else:
        raise NotImplementedError("Dimensionality is constrained to 2.")

    return convolve(x, kernel, fill_value=np.nan, boundary="fill")


def move_rms(x, w, window="box"):
    r"""Calculate the moving root-mean-square of a signal."""
    return np.sqrt(move_avg(x**2, w, window=window))


def move_std(x, w, window="box"):
    r"""Calculate the moving standard deviation of a signal."""
    return np.sqrt(
        move_avg(x**2, w, window=window) - move_avg(x, w, window=window) ** 2
    )
